# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import os
import unittest
from time import time

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from neon_utils.message_utils import *

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

VALID_STRING = 'This is only a test!'
VALID_BYTES = b'This is only a test!'
ENCODED_UTF8 = "VGhpcyBpcyBvbmx5IGEgdGVzdCE="
ENCODED_UTF16 = "䝖灨祣灂祣療浢㕸䝉杅䝤穖䍤㵅"


def get_message_standard(message):
    print(message)
    return dig_for_message()


def get_message_alt_name(msg):
    print(msg)
    return dig_for_message()


def get_message_no_name(_):
    return dig_for_message()


class MessageUtilTests(unittest.TestCase):
    def test_request_from_mobile(self):
        from_mobile = request_from_mobile(Message("", {}, {"mobile": True}))
        self.assertTrue(from_mobile)

        not_from_mobile = request_from_mobile(Message("", {}, {}))
        self.assertFalse(not_from_mobile)

    def test_get_message_username(self):
        with_user = get_message_user(Message("", {}, {"username": "testrunner"}))
        self.assertEqual(with_user, "testrunner")

        without_user = get_message_user(Message(""))
        self.assertIsNone(without_user)

    def test_get_message_username_invalid_arg(self):
        with self.assertRaises(ValueError):
            get_message_user(None)

        with self.assertRaises(AttributeError):
            get_message_user("Nobody")

    def test_encode_bytes_valid(self):
        encoded_8 = encode_bytes_to_b64_string(VALID_BYTES, 'utf-8')
        self.assertIsInstance(encoded_8, str)
        self.assertEqual(encoded_8, ENCODED_UTF8)

    def test_encode_bytes_invalid_data_type(self):
        with self.assertRaises(ValueError):
            encode_bytes_to_b64_string(ENCODED_UTF8)

    def test_encode_bytes_invalid_charset(self):
        with self.assertRaises(EncodingError):
            encode_bytes_to_b64_string(VALID_BYTES, "INVALID_ENCODING")

    def test_decode_bytes_valid(self):
        decoded_8 = decode_b64_string_to_bytes(ENCODED_UTF8, 'utf-8')
        self.assertIsInstance(decoded_8, bytes)
        self.assertEqual(decoded_8, VALID_BYTES)

    def test_decode_bytes_invalid_data_type(self):
        with self.assertRaises(ValueError):
            decode_b64_string_to_bytes(VALID_BYTES)

    def test_decode_bytes_incorrect_encoding(self):
        with self.assertRaises(EncodingError):
            decode_b64_string_to_bytes(ENCODED_UTF16, "utf-8")

    def test_decode_bytes_invalid_encoding(self):
        with self.assertRaises(EncodingError):
            decode_b64_string_to_bytes(ENCODED_UTF8, "INVALID_ENCODING")

    def test_dig_for_message_simple(self):
        test_msg = Message("test message", {"test": "data"}, {"time": time()})
        self.assertEqual(test_msg, get_message_standard(test_msg))
        test_msg = Message("test message", {"test": "data"}, {"time": time()})
        self.assertEqual(test_msg, get_message_alt_name(test_msg))
        test_msg = Message("test message", {"test": "data"}, {"time": time()})
        self.assertEqual(test_msg, get_message_no_name(test_msg))

    def test_dig_for_message_nested(self):
        message = Message("test message", {"test": "data"}, {"time": time()})

        def simple_wrapper():
            return get_message_no_name(message)

        self.assertEqual(simple_wrapper(), message)

        message = Message("test message", {"test": "data"}, {"time": time()})

        def get_message():
            return dig_for_message()

        def wrapper_method(msg):
            self.assertEqual(msg, get_message())

        wrapper_method(message)

    def test_dig_for_message_invalid_type(self):
        tester = Message("test message", {"test": "data"}, {"time": time()})

        def wrapper_method(_):
            return dig_for_message()
        self.assertIsNone(wrapper_method(dict()))

    def test_dig_for_message_no_method_call(self):
        message = Message("test message", {"test": "data"}, {"time": time()})
        self.assertIsNone(dig_for_message())

    def test_resolve_message(self):
        def wrapper_method(message, function: callable,
                           fn_args: list = None, fn_kwargs: dict = None):
            fn_args = fn_args or list()
            fn_kwargs = fn_kwargs or dict()
            function(*fn_args, **fn_kwargs)

        def nested_get_message(message=None):
            get_message_simple(message)
            get_message_multi_args("test", message)

        @resolve_message
        def get_message_simple(message=None):
            self.assertIsInstance(message, Message)
            self.assertEqual(message, test_message)

        @resolve_message
        def get_message_invalid_args(test=None):
            self.assertNotIsInstance(test, Message)

        @resolve_message
        def get_message_kwargs(*args, **kwargs):
            if args:
                self.assertIn(test_message, args)
            else:
                self.assertIsInstance(kwargs["message"], Message)
                self.assertEqual(kwargs["message"], test_message)

        @resolve_message
        def get_message_multi_args(test, message=None):
            self.assertEqual(test, "test")
            self.assertEqual(message, test_message)

        test_message = Message("test", {"data": "val"}, {"context": False})
        wrapper_method(test_message, get_message_simple)
        wrapper_method(test_message, get_message_simple, [test_message])
        wrapper_method(test_message, get_message_simple,
                       fn_kwargs={"message": test_message})

        wrapper_method(test_message, get_message_invalid_args)
        wrapper_method(test_message, get_message_invalid_args, ["test"])

        wrapper_method(test_message, get_message_kwargs)
        wrapper_method(test_message, get_message_kwargs,
                       fn_kwargs={"message": test_message})
        wrapper_method(test_message, get_message_kwargs,
                       fn_args=[test_message])

        wrapper_method(test_message, nested_get_message)

    def test_request_for_neon(self):
        from neon_utils.message_utils import request_for_neon

        minimal_message = Message("test")
        self.assertTrue(request_for_neon(minimal_message))
        self.assertFalse(request_for_neon(minimal_message, ww_enabled=False))

        neon_should_respond = Message("test", {},
                                      {"neon_should_respond": True})
        self.assertTrue(request_for_neon(neon_should_respond))
        self.assertTrue(request_for_neon(neon_should_respond,
                                         ww_enabled=False))

        neon_voc = Message("test", {"neon": "nyan"},
                           {"neon_shoud_respond": False})
        self.assertTrue(request_for_neon(neon_voc, ww_enabled=False))
        self.assertFalse(request_for_neon(neon_voc, "mycroft",
                                          ww_enabled=False))

        server_request_public = Message("test", {},
                                        {"klat_data": {"title": "title"}})
        self.assertFalse(request_for_neon(server_request_public))

        server_request_private = Message("test", {},
                                         {"klat_data": {
                                             "title": "!PRIVATE:user"}})
        self.assertTrue(request_for_neon(server_request_private))

    def test_build_message(self):
        from neon_utils.message_utils import MessageKind, build_message
        base_message = Message("test", context={"context": time()})
        speak = build_message("neon speak", "test utterance", base_message)
        self.assertEqual(speak.context['context'],
                         base_message.context['context'])

        self.assertEqual(speak.msg_type, "speak")
        self.assertEqual(speak.data, {"utterance": "test utterance",
                                      "lang": "en-US",
                                      "speaker": {
                                          "name": "Neon",
                                          "language": "en-us",
                                          "gender": "female",
                                          "voice": "",
                                          "override_user": True
                                      }})

        utt = "testing again"
        speaker = {"name": "test name",
                   "lang": "uk-ua"}
        base_message.data['lang'] = 'uk-ua'
        execute = build_message(MessageKind.EXECUTE, utt, base_message,
                                speaker)
        self.assertEqual(execute.msg_type, "skills:execute.utterance")
        self.assertEqual(execute.data["utterances"], [utt])
        self.assertEqual(execute.data["lang"], "uk-ua")
        self.assertEqual(execute.data["speaker"], speaker)
        self.assertEqual(execute.context["cc_data"],
                         {"request": utt,
                          "emit_response": True,
                          "execute_from_script": True
                          })

    def test_neon_must_respond(self):
        from neon_utils.message_utils import neon_must_respond
        self.assertFalse(neon_must_respond())
        private_message_solo = Message("", {},
                                       {"klat_data": {
                                           "title": "!PRIVATE:user"}})
        private_message_neon = Message("", {},
                                       {"klat_data": {
                                           "title": "!PRIVATE:user,Neon"}})
        private_message_neon_plus = \
            Message("", {},
                    {"klat_data": {"title": "!PRIVATE:user,Neon,user1"}})
        public_message = Message("", {},
                                 {"klat_data": {
                                     "title": "Test Conversation"}})
        first_message = Message("",
                                {"utterance": "Welcome to your private "
                                              "conversation with Neon"},
                                {"klat_data": {
                                    "title": "!PRIVATE:user"}})
        self.assertFalse(neon_must_respond())
        self.assertTrue(neon_must_respond(private_message_solo))
        self.assertTrue(neon_must_respond(private_message_neon))
        self.assertFalse(neon_must_respond(private_message_neon_plus))
        self.assertFalse(neon_must_respond(public_message))
        self.assertFalse(neon_must_respond(first_message))


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from ovos_utils import LOG

from neon_utils.skills.neon_skill import NeonSkill
from ovos_utils.intents import IntentLayers
from ovos_workshop.decorators.layers import IntentLayers
from ovos_workshop.skills.fallback import FallbackSkillV1


# TODO: Consider deprecation and implementing ovos_workshop directly
class NeonFallbackSkill(FallbackSkillV1, NeonSkill):
    """
    Class that extends the NeonSkill and FallbackSkill classes to provide
    NeonSkill functionality to any Fallback skill subclassing this class.
    """
    def __init__(self, *args, **kwargs):
        # Manual init of OVOSSkill
        self.private_settings = None
        self._threads = []
        self._original_converse = self.converse
        self.intent_layers = IntentLayers()
        self.audio_service = None

        # Manual init of FallbackSkill
        #  list of fallback handlers registered by this instance
        self.instance_fallback_handlers = []
        NeonSkill.__init__(self, *args, **kwargs)
        LOG.debug(f"instance_handlers={self.instance_fallback_handlers}")
        LOG.debug(f"class_handlers={FallbackSkillV1.fallback_handlers}")

    @property
    def fallback_config(self):
        # "skill_id": priority (int)  overrides
        return self.config_core["skills"].get("fallbacks", {})

    @classmethod
    def _register_fallback(cls, *args, **kwargs):
        LOG.debug(f"register fallback")
        FallbackSkillV1._register_fallback(*args, **kwargs)

    def _register_decorated(self):
        # Explicitly overridden to ensure the correct super call is made
        LOG.debug(f"Registering decorated methods for {self.skill_id}")
        try:
            FallbackSkillV1._register_decorated(self)
        except Exception as e:
            LOG.error(e)
            NeonSkill._register_decorated(self)
            from ovos_utils.skills import get_non_properties
            for attr_name in get_non_properties(self):
                method = getattr(self, attr_name)
                if hasattr(method, 'fallback_priority'):
                    self.register_fallback(method, method.fallback_priority)

    def register_fallback(self, *args, **kwargs):
        LOG.debug(f"Registering fallback handler for {self.skill_id}")
        FallbackSkillV1.register_fallback(self, *args, **kwargs)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
import os
import shutil
import sys
import unittest
import pytest

from multiprocessing import Event
from os.path import join
from threading import Thread
from time import sleep, time

from ovos_bus_client import Message
from ovos_utils.messagebus import FakeBus
from mock import Mock

from neon_utils.skills import NeonSkill, CommonMessageSkill, CommonPlaySkill, CommonQuerySkill, NeonFallbackSkill

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from neon_utils.cache_utils import LRUCache
from neon_utils.signal_utils import check_for_signal
sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from skills import (PatchedMycroftSkill, TestCMS, TestCPS, TestCQS, TestFBS,
                    TestPatchedSkill, TestInstructorSkill, TestChatSkill,
                    TestNeonSkill, TestKioskSkill, TestMycroftFallbackSkill)

MycroftSkill = PatchedMycroftSkill
ROOT_DIR = os.path.dirname(os.path.realpath(__file__))
SKILL_PATH = os.path.join(ROOT_DIR, "skills")
BUS = FakeBus()


def get_test_mycroft_skill(bus_events: dict):
    skill = MycroftSkill()
    bus = FakeBus()
    for event, callback in bus_events.items():
        bus.on(event, callback)
    bus.run_in_thread()
    if hasattr(skill, "_startup"):
        skill._startup(bus)
    else:
        skill.bind(bus)
    return skill


def get_test_neon_skill(bus_events: dict):
    skill = NeonSkill()
    bus = FakeBus()
    for event, callback in bus_events.items():
        bus.on(event, callback)
    bus.run_in_thread()
    if hasattr(skill, "_startup"):
        skill._startup(bus)
    else:
        skill.bind(bus)
    return skill


def create_skill(skill):
    skill = skill()
    if hasattr(skill, "_startup"):
        skill._startup(BUS)
    return skill


class SkillObjectTests(unittest.TestCase):
    # def test_alt_fallback_skill(self):
    #     skill = create_skill(TestMycroftFallbackSkill)
    #     self.assertIsInstance(skill, MycroftSkill)
    #     self.assertIsInstance(skill, FallbackSkill)

    def test_common_message_skill_init(self):
        skill = create_skill(TestCMS)
        # self.assertIsInstance(skill, MycroftSkill)
        self.assertIsInstance(skill, NeonSkill)
        self.assertIsInstance(skill, CommonMessageSkill)
        self.assertEqual(skill.name, "Test Common Message Skill")

    def test_common_play_skill_init(self):
        skill = create_skill(TestCPS)
        # self.assertIsInstance(skill, MycroftSkill)
        self.assertIsInstance(skill, NeonSkill)
        self.assertIsInstance(skill, CommonPlaySkill)
        self.assertEqual(skill.name, "Test Common Play Skill")

    def test_common_query_skill_init(self):
        skill = create_skill(TestCQS)
        # self.assertIsInstance(skill, MycroftSkill)
        self.assertIsInstance(skill, NeonSkill)
        self.assertIsInstance(skill, CommonQuerySkill)
        self.assertEqual(skill.name, "Test Common Query Skill")

    def test_fallback_skill_init(self):
        skill = create_skill(TestFBS)
        # self.assertIsInstance(skill, MycroftSkill)
        self.assertIsInstance(skill, NeonSkill)
        self.assertIsInstance(skill, NeonFallbackSkill)
        # self.assertIsInstance(skill, FallbackSkill)
        self.assertEqual(skill.name, "Test Fallback Skill")

    def test_neon_skill_init(self):
        skill = create_skill(TestNeonSkill)
        # self.assertIsInstance(skill, MycroftSkill)
        self.assertIsInstance(skill, NeonSkill)
        self.assertEqual(skill.name, "Test Neon Skill")

        self.assertIsInstance(skill.lru_cache, LRUCache)
        self.assertIsInstance(skill.sys_tz, datetime.tzinfo)
        self.assertIsInstance(skill.gui_enabled, bool)
        self.assertIsInstance(skill.neon_core, bool)
        self.assertIsInstance(skill.actions_to_confirm, dict)

        self.assertIsInstance(skill.skill_mode, str)
        self.assertIsInstance(skill.extension_time, int)

        if skill.lang_detector:
            from ovos_plugin_manager.templates.language import LanguageDetector
            self.assertIsInstance(skill.lang_detector, LanguageDetector)
        if skill.translator:
            from ovos_plugin_manager.templates.language import LanguageTranslator
            self.assertIsInstance(skill.translator, LanguageTranslator)

        self.assertIsInstance(skill.settings, dict)
        self.assertIsInstance(skill.location_timezone, str)
        self.assertIsInstance(skill.preference_skill(), dict)
        self.assertEqual(skill.settings, skill.preference_skill())
        self.assertIsInstance(skill.file_system.path, str)
        # self.assertIsNotNone(skill.settings_meta)
        # self.assertEqual(skill.file_system.path, skill.settings_write_path)
        # self.assertNotEqual(os.path.basename(skill.file_system.path),
        #                     skill.name)
        expected_methods = ['init_dialog', 'make_active', 'translate',
                            'translate_namedvalues', 'translate_list',
                            'translate_template', 'bind',
                            'handle_settings_change', 'detach',
                            'get_intro_message', 'converse', 'get_response',
                            'ask_yesno', 'ask_selection', 'voc_match',
                            'report_metric', 'send_email',
                            'register_resting_screen', 'find_resource',
                            'add_event', 'remove_event', 'register_intent',
                            'register_intent_file', 'register_entity_file',
                            'handle_enable_intent', 'handle_disable_intent',
                            'disable_intent', 'enable_intent', 'set_context',
                            'remove_context', 'handle_set_cross_context',
                            'handle_remove_cross_context',
                            'set_cross_skill_context',
                            'remove_cross_skill_context', 'register_vocabulary',
                            'register_regex', 'speak', 'speak_dialog',
                            'load_dialog_files', 'load_data_files',
                            'load_vocab_files', 'load_regex_files', 'stop',
                            'shutdown', 'default_shutdown', 'schedule_event',
                            'schedule_repeating_event',
                            'update_scheduled_event', 'cancel_scheduled_event',
                            'get_scheduled_event_status',
                            'cancel_all_repeating_events']
        for method in expected_methods:
            self.assertTrue(hasattr(skill, method), method)

    def test_patched_mycroft_skill_init(self):
        skill = create_skill(TestPatchedSkill)
        self.assertIsInstance(skill, MycroftSkill)
        self.assertEqual(skill.name, "Test Mycroft Skill")

        # self.assertEqual(skill.file_system.path, skill.settings_write_path)
        # self.assertNotEqual(os.path.basename(skill.file_system.path), skill.name)

    def test_instructor_skill_init(self):
        skill = create_skill(TestInstructorSkill)
        # self.assertIsInstance(skill, MycroftSkill)
        self.assertEqual(skill.name, "Test Instructor Skill")


class KioskSkillTests(unittest.TestCase):
    skill = create_skill(TestKioskSkill)

    def setUp(self) -> None:
        self.skill.setup_new_interaction = Mock(return_value=True)
        self.skill.handle_new_interaction = Mock()
        self.skill.handle_end_interaction = Mock()
        self.skill.handle_user_utterance = Mock()
        self.skill.speak_dialog = Mock()
        self.skill.event_scheduler.schedule_event = Mock()

    def test_kiosk_skill_init(self):
        # self.assertIsInstance(self.skill, MycroftSkill)
        self.assertIsInstance(self.skill, NeonSkill)
        self.assertEqual(self.skill.greeting_dialog, 'greeting')
        self.assertEqual(self.skill.goodbye_dialog, 'goodbye')
        self.assertEqual(self.skill.timeout_dialog, 'timeout')
        self.assertEqual(self.skill.error_dialog, 'error')

    def test_skill_timeout(self):
        test_message = Message('test', {}, {'username': 'test_user'})

        # Test Timeout
        self.skill._timeout_seconds = 1
        self.skill.start_interaction(test_message)
        self.assertEqual(set(self.skill._active_users.keys()), {'test_user'})
        self.skill.setup_new_interaction.assert_called_once_with(test_message)
        self.skill.handle_new_interaction.assert_called_once_with(test_message)
        self.skill.speak_dialog.assert_called_once_with(self.skill.greeting_dialog)
        args = self.skill.event_scheduler.schedule_event.call_args
        self.assertEqual(args[0][0], self.skill._handle_timeout)
        self.assertIsInstance(args[0][1], datetime.datetime)
        self.assertEqual(args[0][2], test_message.data)
        self.assertEqual(args[0][3], f"{self.skill.skill_id}:timeout_test_user")
        self.assertEqual(args[1]['context'], {**test_message.context,
                                              **{'skill_id': self.skill.skill_id}})
        expiration_message = Message(f'{self.skill.skill_id}:timeout_test_user',
                                     test_message.data, args[1]['context'])
        self.skill._handle_timeout(expiration_message)
        self.skill.speak_dialog.assert_called_with(self.skill.timeout_dialog)
        self.assertEqual(self.skill._active_users, dict())

    def test_skill_error_handling(self):
        test_message = Message('test', {}, {'username': 'test_user'})

        # Test error handling
        self.skill.handle_error(test_message)
        self.skill.speak_dialog.assert_called_once_with(self.skill.error_dialog,
                                                        message=test_message)

    def test_skill_setup_failure(self):
        test_message = Message('test', {}, {'username': 'test_user'})

        # Test setup failure
        self.skill.setup_new_interaction.return_value = False
        self.skill.start_interaction(test_message)
        self.skill.setup_new_interaction.assert_called_once_with(test_message)
        self.assertEqual(self.skill._active_users, dict())

    def test_skill_normal_interaction(self):
        test_message = Message('test', {}, {'username': 'test_user'})

        self.skill._timeout_seconds = 30
        self.skill.start_interaction(test_message)
        self.assertEqual(set(self.skill._active_users.keys()), {'test_user'})
        self.skill.setup_new_interaction.assert_called_once_with(test_message)
        self.skill.handle_new_interaction.assert_called_once_with(test_message)
        self.skill.speak_dialog.assert_called_once_with(self.skill.greeting_dialog)

        # Test simple converse
        sleep(1)
        converse_message = Message('converse', {}, {'username': 'test_user'})
        self.assertTrue(self.skill.converse(converse_message))
        self.skill.handle_user_utterance.assert_called_once_with(converse_message)
        self.assertAlmostEqual(self.skill._active_users['test_user'], time(), 1)

        # Test long-running converse handler
        msg = None
        event = Event()

        def _handle_utterance(message):
            nonlocal msg
            msg = message
            sleep(5)
            event.set()

        self.skill.handle_user_utterance = _handle_utterance
        self.assertTrue(self.skill.converse(converse_message))
        self.assertFalse(event.is_set())
        event.wait(5)
        finished_time = time()
        self.assertEqual(msg, converse_message)
        sleep(0.5)
        self.assertAlmostEqual(self.skill._active_users['test_user'],
                               finished_time, 1)

        # Test end interaction
        end_message = Message('end', {}, {'username': 'test_user'})
        self.skill.end_interaction(end_message)
        self.skill.handle_end_interaction.assert_called_once_with(end_message)
        self.skill.speak_dialog.assert_called_with(self.skill.goodbye_dialog)
        self.assertEqual(self.skill._active_users, dict())

    def test_skill_stop(self):
        real_converse = self.skill._handle_converse
        self.skill._handle_converse = Mock()

        test_message = Message('test', {"utterance": "stop"}, {'username': 'test_user'})

        self.skill._timeout_seconds = 30
        self.skill.start_interaction(test_message)
        self.assertEqual(set(self.skill._active_users.keys()), {'test_user'})

        self.skill.converse(test_message)
        self.skill._handle_converse.assert_not_called()
        self.assertNotIn("test_user", self.skill._active_users.keys())

        self.skill._handle_converse = real_converse


class ChatSkillTests(unittest.TestCase):
    def test_skill_init(self):
        msg: Message = None

        def handle_register(message):
            nonlocal msg
            msg = message

        BUS.once("register_chat_handler", handle_register)

        # Test decorator registration
        skill = create_skill(TestChatSkill)
        self.assertIsInstance(msg, Message)
        self.assertEqual(msg.data['name'], "Test Bot")
        self.assertEqual(msg.context['skill_id'], skill.skill_id)

        # Test decorated method
        test_message = Message("chat.Test Bot", {"test_response": "nothing"},
                               {'test': True})
        resp = BUS.wait_for_response(test_message)
        self.assertEqual(resp.msg_type, f'{test_message.msg_type}.response')
        self.assertEqual(resp.data,
                         {'response': test_message.data['test_response']})
        self.assertTrue(resp.context['test'])
        self.assertEqual(resp.context['skill_id'], skill.skill_id)


class PatchedMycroftSkillTests(unittest.TestCase):
    def test_get_response_simple(self):
        def handle_speak(_):
            check_for_signal("isSpeaking")
            spoken.set()

        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog)
            test_results[idx] = resp

        test_results = dict()
        spoken = Event()
        test_dialog = "testing get response."
        message = Message("recognizer_loop:utterance",
                          {"utterances": ["testing one", "testing 1",
                                          "resting one"]},
                          {"timing": {},
                           "username": "local"})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        t = Thread(target=skill_response_thread,
                   args=(skill, message.context["username"]), daemon=True)
        t.start()
        spoken.wait(30)
        sleep(1)
        skill.converse(message)
        t.join(5)
        self.assertEqual(test_results[message.context["username"]],
                         message.data["utterances"][0])

    def test_get_response_interrupt_prompt(self):
        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog)
            test_results[idx] = resp

        test_results = dict()
        spoken = Event()
        test_dialog = "testing get response."
        message = Message("recognizer_loop:utterance",
                          {"utterances": ["testing one", "testing 1",
                                          "resting one"]},
                          {"timing": {},
                           "username": "local"})

        skill = get_test_mycroft_skill({})
        t = Thread(target=skill_response_thread,
                   args=(skill, message.context["username"]), daemon=True)
        t.start()
        sleep(1)
        skill.converse(message)
        t.join(5)
        self.assertEqual(test_results[message.context["username"]],
                         message.data["utterances"][0])

    def test_get_response_no_username(self):
        def handle_speak(_):
            check_for_signal("isSpeaking")
            spoken.set()

        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog)
            test_results[idx] = resp

        test_results = dict()
        spoken = Event()
        test_dialog = "testing get response."
        message = Message("recognizer_loop:utterance",
                          {"utterances": ["testing one", "testing 1",
                                          "resting one"]},
                          {"timing": {}})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        t = Thread(target=skill_response_thread, args=(skill, "0"),
                   daemon=True)
        t.start()
        spoken.wait(30)
        sleep(1)
        skill.converse(message)
        t.join(5)
        self.assertEqual(test_results["0"], message.data["utterances"][0])

    def test_get_response_multi_user(self):
        def handle_speak(_):
            check_for_signal("isSpeaking")
            spoken.set()

        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog, message=Message(
                "converse_message", {}, {"username": "valid_converse_user"}))
            test_results[idx] = resp

        test_results = dict()
        spoken = Event()
        test_dialog = "testing get response multi user."
        valid_message = Message("recognizer_loop:utterance",
                                {"utterances": ["testing one",
                                                "testing 1", "resting one"]},
                                {"timing": {},
                                 "username": "valid_converse_user"})
        invalid_message = Message("recognizer_loop:utterance",
                                  {"utterances": ["invalid return"]},
                                  {"timing": {},
                                   "username": "invalid_converse_user"})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        t = Thread(target=skill_response_thread,
                   args=(skill, valid_message.context["username"]),
                   daemon=True)
        t.start()
        spoken.wait(30)
        sleep(1)
        skill.converse(invalid_message)
        skill.converse(valid_message)
        skill.converse(invalid_message)
        t.join(5)
        self.assertEqual(test_results[valid_message.context["username"]],
                         valid_message.data["utterances"][0])

    def test_get_response_dig_for_message(self):
        def handle_speak(_):
            check_for_signal("isSpeaking")
            spoken.set()

        def skill_response_thread(s: MycroftSkill, idx: str):
            def intent_handler(message):
                resp = s.get_response(test_dialog)
                test_results[idx] = resp
            intent_handler(Message("converse_message",
                                   {}, {"username": "valid_converse_user"}))

        test_results = dict()
        spoken = Event()
        test_dialog = "testing get response multi user."
        valid_message = Message("recognizer_loop:utterance",
                                {"utterances": ["testing one", "testing 1",
                                                "resting one"]},
                                {"timing": {},
                                 "username": "valid_converse_user"})
        invalid_message = Message("recognizer_loop:utterance",
                                  {"utterances": ["invalid return"]},
                                  {"timing": {},
                                   "username": "invalid_converse_user"})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        t = Thread(target=skill_response_thread,
                   args=(skill, valid_message.context["username"]),
                   daemon=True)
        t.start()
        spoken.wait(30)
        sleep(1)
        skill.converse(invalid_message)
        skill.converse(valid_message)
        skill.converse(invalid_message)
        t.join(5)
        self.assertEqual(test_results[valid_message.context["username"]],
                         valid_message.data["utterances"][0])

    def test_get_response_no_response(self):
        def handle_speak(_):
            check_for_signal("isSpeaking")
            spoken.set()

        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog, num_retries=0, message=Message(
                "converse_message", {}, {"username": "valid_converse_user"}))
            test_results[idx] = resp

        test_results = dict()
        spoken = Event()
        test_dialog = "testing get response multi user."
        valid_message = Message("recognizer_loop:utterance",
                                {"utterances": ["testing one", "testing 1",
                                                "resting one"]},
                                {"timing": {},
                                 "username": "valid_converse_user"})
        invalid_message = Message("recognizer_loop:utterance",
                                  {"utterances": ["invalid return"]},
                                  {"timing": {},
                                   "username": "invalid_converse_user"})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        skill._get_response_timeout = 5
        t = Thread(target=skill_response_thread,
                   args=(skill, valid_message.context["username"]),
                   daemon=True)
        t.start()
        self.assertTrue(spoken.wait(30))
        sleep(1)
        skill.converse(invalid_message)
        t.join(30)
        self.assertIsNone(test_results[valid_message.context["username"]])

    def test_get_response_retried_no_dialog(self):
        def handle_speak(_):
            check_for_signal("isSpeaking")
            spoken.set()

        def handle_failed(_):
            failed.set()

        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog, num_retries=1,
                                  on_fail=handle_failed, message=Message(
                "converse_message", {}, {"username": "valid_converse_user"}))
            test_results[idx] = resp

        test_results = dict()
        spoken = Event()
        failed = Event()
        test_dialog = "testing get response multi user."
        valid_message = Message("recognizer_loop:utterance",
                                {"utterances": ["testing one", "testing 1",
                                                "resting one"]},
                                {"timing": {},
                                 "username": "valid_converse_user"})
        invalid_message = Message("recognizer_loop:utterance",
                                  {"utterances": ["invalid return"]},
                                  {"timing": {},
                                   "username": "invalid_converse_user"})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        skill._get_response_timeout = 5
        t = Thread(target=skill_response_thread,
                   args=(skill, valid_message.context["username"]),
                   daemon=True)
        t.start()

        self.assertTrue(spoken.wait(30))
        sleep(1)  # TODO: This is patching a race condition in _wait_response
        skill.converse(invalid_message)
        failed.wait(5)
        sleep(1)  # TODO: This is patching a race condition in _wait_response
        skill.converse(valid_message)

        self.assertTrue(failed.is_set())
        t.join(30)
        self.assertEqual(test_results[valid_message.context["username"]],
                         valid_message.data["utterances"][0])

    def test_get_response_retried_with_dialog(self):
        def handle_speak(_):
            check_for_signal("isSpeaking")
            spoken.set()

        def handle_failed(_):
            failed.set()
            return "test speech"

        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog, num_retries=1,
                                  on_fail=handle_failed, message=Message(
                "converse_message", {}, {"username": "valid_converse_user"}))
            test_results[idx] = resp

        test_results = dict()
        spoken = Event()
        failed = Event()
        test_dialog = "testing get response multi user."
        valid_message = Message("recognizer_loop:utterance",
                                {"utterances": ["testing one", "testing 1",
                                                "resting one"]},
                                {"timing": {},
                                 "username": "valid_converse_user"})
        invalid_message = Message("recognizer_loop:utterance",
                                  {"utterances": ["invalid return"]},
                                  {"timing": {},
                                   "username": "invalid_converse_user"})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        skill._get_response_timeout = 5
        t = Thread(target=skill_response_thread,
                   args=(skill, valid_message.context["username"]),
                   daemon=True)
        t.start()

        self.assertTrue(spoken.wait(30))
        spoken.clear()
        sleep(1)  # TODO: This is patching a race condition in _wait_response
        skill.converse(invalid_message)
        failed.wait(5)
        spoken.wait(30)
        sleep(1)  # TODO: This is patching a race condition in _wait_response
        skill.converse(valid_message)

        self.assertTrue(failed.is_set())
        t.join(30)
        self.assertEqual(test_results[valid_message.context["username"]],
                         valid_message.data["utterances"][0])

    def test_get_response_validator_return_utterance(self):
        # returns utterance if validotor return value is True
        def handle_speak(_):
            check_for_signal("isSpeaking")
            spoken.set()

        def is_valid(_):
            test_results["validator"] = True
            return True

        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog, validator=is_valid,
                                  message=Message(
                                      "converse_message", {},
                                      {"username": "valid_converse_user"}))
            test_results[idx] = resp

        test_results = dict()
        spoken = Event()
        test_dialog = "testing get response multi user."
        valid_message = Message("recognizer_loop:utterance",
                                {"utterances": ["testing one", "testing 1",
                                                "resting one"]},
                                {"timing": {},
                                 "username": "valid_converse_user"})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        t = Thread(target=skill_response_thread,
                   args=(skill, valid_message.context["username"]),
                   daemon=True)
        t.start()
        spoken.wait(30)
        sleep(1)
        skill.converse(valid_message)
        t.join(30)
        self.assertTrue(test_results["validator"])
        self.assertEqual(test_results[valid_message.context["username"]],
                         valid_message.data["utterances"][0])

    def test_get_response_validator_return_value_1(self):
        def handle_speak(_):
            check_for_signal("isSpeaking")
            spoken.set()

        def is_valid(_):
            test_results["validator"] = True
            return valid_return

        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog, validator=is_valid,
                                  message=Message(
                                      "converse_message", {},
                                      {"username": "valid_converse_user"}))
            test_results[idx] = resp

        test_results = dict()
        valid_return = datetime.datetime(2022,11,29,0,0)
        spoken = Event()
        test_dialog = "testing get response multi user."
        valid_message = Message("recognizer_loop:utterance",
                                {"utterances": ["today"]},
                                {"timing": {},
                                 "username": "valid_converse_user"})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        t = Thread(target=skill_response_thread,
                   args=(skill, valid_message.context["username"]),
                   daemon=True)
        t.start()
        spoken.wait(30)
        sleep(1)
        skill.converse(valid_message)
        t.join(30)
        self.assertTrue(test_results["validator"])
        self.assertEqual(test_results[valid_message.context["username"]],
                         valid_return)

    def test_get_response_validator_return_value_2(self):
        # validator should also accept faulty values that ar not None/False
        def handle_speak(_):
            check_for_signal("isSpeaking")
            spoken.set()

        def is_valid(_):
            test_results["validator"] = True
            return valid_return

        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog, validator=is_valid,
                                  message=Message(
                                      "converse_message", {},
                                      {"username": "valid_converse_user"}))
            test_results[idx] = resp

        test_results = dict()
        valid_return = 0
        spoken = Event()
        test_dialog = "testing get response multi user."
        valid_message = Message("recognizer_loop:utterance",
                                {"utterances": ["got zero items"]},
                                {"timing": {},
                                 "username": "valid_converse_user"})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        t = Thread(target=skill_response_thread,
                   args=(skill, valid_message.context["username"]),
                   daemon=True)
        t.start()
        spoken.wait(30)
        sleep(1)
        skill.converse(valid_message)
        t.join(30)
        self.assertTrue(test_results["validator"])
        self.assertEqual(test_results[valid_message.context["username"]],
                         valid_return)

    def test_get_response_validator_fail(self):
        def handle_speak(_):
            check_for_signal("isSpeaking")
            spoken.set()

        def is_valid(_):
            test_results["validator"] = True
            return False

        on_fail = Mock()

        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog, validator=is_valid,
                                  on_fail=on_fail, message=Message(
                    "converse_message", {},
                    {"username": "valid_converse_user"}))
            test_results[idx] = resp

        test_results = dict()
        spoken = Event()
        test_dialog = "testing get response multi user."
        valid_message = Message("recognizer_loop:utterance",
                                {"utterances": ["testing one", "testing 1",
                                                "resting one"]},
                                {"timing": {},
                                 "username": "valid_converse_user"})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        t = Thread(target=skill_response_thread,
                   args=(skill, valid_message.context["username"]),
                   daemon=True)
        t.start()
        spoken.wait(30)
        sleep(1)
        skill.converse(valid_message)
        t.join(30)
        self.assertTrue(test_results["validator"])
        on_fail.assert_called_once()
        on_fail.assert_called_with("testing one")

    def test_get_response_klat(self):
        speak_message: Message = None
        def handle_speak(msg):
            nonlocal speak_message
            speak_message = msg
            spoken.set()

        def skill_response_thread(s: MycroftSkill, idx: str):
            resp = s.get_response(test_dialog, message=Message(
                "converse_message", {},
                {"username": "valid_converse_user",
                 "klat_data": {"title": "Public Convo"}}))
            test_results[idx] = resp

        test_results = dict()
        spoken = Event()
        test_dialog = "testing get response multi user."
        valid_message = Message("recognizer_loop:utterance",
                                {"utterances": ["testing one",
                                                "testing 1", "resting one"]},
                                {"timing": {},
                                 "username": "valid_converse_user",
                                 "klat_data": {"title": "Public Convo"}})
        invalid_message = Message("recognizer_loop:utterance",
                                  {"utterances": ["invalid return"]},
                                  {"timing": {},
                                   "username": "invalid_converse_user",
                                   "klat_data": {"title": "Public Convo"}})

        skill = get_test_mycroft_skill({"speak": handle_speak})
        t = Thread(target=skill_response_thread,
                   args=(skill, valid_message.context["username"]),
                   daemon=True)
        t.start()
        spoken.wait(30)
        self.assertIsInstance(speak_message, Message)
        self.assertTrue(speak_message.data['utterance'].startswith('@valid_converse_user '),
                        speak_message.data['utterance'])
        sleep(1)
        skill.converse(invalid_message)
        skill.converse(valid_message)
        skill.converse(invalid_message)
        t.join(5)
        self.assertEqual(test_results[valid_message.context["username"]],
                         valid_message.data["utterances"][0])

# TODO: test get_response with `speak_bus_api`

    def test_speak_simple_valid(self):
        handle_speak = Mock()
        utterance = "test to speak"
        skill = get_test_mycroft_skill({"speak": handle_speak})
        skill.speak(utterance)
        handle_speak.assert_called_once()
        msg = handle_speak.call_args.args[0]
        self.assertIsInstance(msg, Message)
        self.assertEqual(msg.data["utterance"], utterance)
        self.assertEqual(msg.data["expect_response"], False)
        self.assertIsInstance(msg.data["meta"], dict)
        self.assertIsNone(msg.data["speaker"])
        self.assertEqual(msg.context['destination'], ['audio'])
        self.assertEqual(msg.context['source'], ['skills'])

    def test_speak_speaker_valid(self):
        handle_speak = Mock()
        utterance = "test to speak"
        speaker = {"speaker": "Test Speaker",
                   "language": "en-au",
                   "gender": "female"}
        skill = get_test_mycroft_skill({"speak": handle_speak})
        skill.speak(utterance, speaker=speaker)
        handle_speak.assert_called_once()
        msg = handle_speak.call_args.args[0]
        self.assertIsInstance(msg, Message)
        self.assertEqual(msg.data["utterance"], utterance)
        self.assertEqual(msg.data["expect_response"], False)
        self.assertIsInstance(msg.data["meta"], dict)
        self.assertEqual(msg.data["speaker"], speaker)
        self.assertEqual(msg.context['destination'], ['audio'])
        self.assertEqual(msg.context['source'], ['skills'])

    def test_speak_simple_with_message_valid(self):
        message = Message("date-time.neon:handle_query_time",
                          {'intent_type': 'date-time.neon:handle_query_time',
                           'target': None,
                           'confidence': 0.6666666666666666,
                           'utterance': 'what time is it neon'},
                          {'client_name': 'mycroft_cli',
                           'source': ['skills'],
                           'destination': 'debug_cli',
                           'client': 'local',
                           'neon_should_respond': False,
                           'timing': {'transcribed': 1631062887.5719671,
                                      'text_parsers': 0.34954047203063965,
                                      'speech_start': 1631062888.1001909},
                           'audio_file': '',
                           'skill_id': 'date-time.neon'})
        handle_speak = Mock()
        utterance = "test to speak"
        skill = get_test_mycroft_skill({"speak": handle_speak})
        skill.speak(utterance, message=message)
        handle_speak.assert_called_once()
        msg = handle_speak.call_args.args[0]
        self.assertIsInstance(msg, Message)
        self.assertEqual(msg.data["utterance"], utterance)
        self.assertEqual(msg.data["expect_response"], False)
        self.assertIsInstance(msg.data["meta"], dict)
        self.assertIsNone(msg.data["speaker"])
        self.assertEqual(msg.context.pop('destination'), ['audio'])
        self.assertEqual(msg.context.pop('source'), ['skills'])
        message.context.pop('source')
        message.context.pop('destination')
        self.assertEqual(message.context, msg.context)

    def test_speak_speaker_with_message_override_valid(self):
        message = Message("date-time.neon:handle_query_time",
                          {'intent_type': 'date-time.neon:handle_query_time',
                           'target': None,
                           'confidence': 0.6666666666666666,
                           'utterance': 'what time is it neon',
                           'speaker': {"speaker": "invalid speaker",
                                       "language": "es-es"}},
                          {'client_name': 'mycroft_cli',
                           'source': ['skills'],
                           'destination': 'debug_cli',
                           'client': 'local',
                           'neon_should_respond': False,
                           'timing': {'transcribed': 1631062887.5719671,
                                      'text_parsers': 0.34954047203063965,
                                      'speech_start': 1631062888.1001909},
                           'audio_file': '',
                           'skill_id': 'date-time.neon'})
        handle_speak = Mock()
        utterance = "test to speak"
        speaker = {"speaker": "Test Speaker",
                   "language": "en-au",
                   "gender": "female"}
        skill = get_test_mycroft_skill({"speak": handle_speak})
        skill.speak(utterance, speaker=speaker, message=message)
        handle_speak.assert_called_once()
        msg = handle_speak.call_args.args[0]
        self.assertIsInstance(msg, Message)
        self.assertEqual(msg.data["utterance"], utterance)
        self.assertEqual(msg.data["expect_response"], False)
        self.assertIsInstance(msg.data["meta"], dict)
        self.assertEqual(msg.data["speaker"], speaker)
        self.assertEqual(msg.context.pop('destination'), ['audio'])
        self.assertEqual(msg.context.pop('source'), ['skills'])
        message.context.pop('source')
        message.context.pop('destination')
        self.assertEqual(message.context, msg.context)

    def test_speak_speaker_with_message_valid(self):
        speaker = {"speaker": "Test Speaker",
                   "language": "en-au",
                   "gender": "female"}
        message = Message("date-time.neon:handle_query_time",
                          {'intent_type': 'date-time.neon:handle_query_time',
                           'target': None,
                           'confidence': 0.6666666666666666,
                           'utterance': 'what time is it neon',
                           'speaker': speaker},
                          {'client_name': 'mycroft_cli',
                           'source': ['skills'],
                           'destination': 'debug_cli',
                           'client': 'local',
                           'neon_should_respond': False,
                           'timing': {'transcribed': 1631062887.5719671,
                                      'text_parsers': 0.34954047203063965,
                                      'speech_start': 1631062888.1001909},
                           'audio_file': '',
                           'skill_id': 'date-time.neon'})
        handle_speak = Mock()
        utterance = "test to speak"

        skill = get_test_mycroft_skill({"speak": handle_speak})
        skill.speak(utterance, message=message)
        handle_speak.assert_called_once()
        msg = handle_speak.call_args.args[0]
        self.assertIsInstance(msg, Message)
        self.assertEqual(msg.data["utterance"], utterance)
        self.assertEqual(msg.data["expect_response"], False)
        self.assertIsInstance(msg.data["meta"], dict)
        self.assertEqual(msg.data["speaker"], speaker)
        self.assertEqual(msg.context.pop('destination'), ['audio'])
        self.assertEqual(msg.context.pop('source'), ['skills'])
        message.context.pop('source')
        message.context.pop('destination')
        self.assertEqual(message.context, msg.context)

    def test_speak_emit_response_valid(self):
        message = Message("date-time.neon:handle_query_time",
                          {'intent_type': 'date-time.neon:handle_query_time',
                           'target': None,
                           'confidence': 0.6666666666666666,
                           'utterance': 'what time is it neon'},
                          {'client_name': 'mycroft_cli',
                           'source': ['skills'],
                           'destination': 'debug_cli',
                           'client': 'local',
                           'neon_should_respond': False,
                           'timing': {'transcribed': 1631062887.5719671,
                                      'text_parsers': 0.34954047203063965,
                                      'speech_start': 1631062888.1001909},
                           'audio_file': '',
                           'skill_id': 'date-time.neon',
                           "cc_data": {"emit_response": True}})
        handle_execute_response = Mock()
        utterance = "test to speak"

        skill = get_test_mycroft_skill(
            {"skills:execute.response": handle_execute_response})
        skill.speak(utterance, message=message)
        handle_execute_response.assert_called_once()
        msg = handle_execute_response.call_args.args[0]
        self.assertIsInstance(msg, Message)
        self.assertEqual(msg.data["utterance"], utterance)
        self.assertEqual(msg.data["expect_response"], False)
        self.assertIsInstance(msg.data["meta"], dict)
        self.assertEqual(msg.context.pop('destination'), ['skills'])
        self.assertEqual(msg.context.pop('source'), ['skills'])
        message.context.pop('source')
        message.context.pop('destination')
        self.assertEqual(message.context, msg.context)

    def test_speak_wait(self):
        from neon_utils.signal_utils import create_signal, check_for_signal, \
            wait_for_signal_clear
        create_signal("neon_speak_api")
        message: Message = None

        def on_speak(msg):
            def handler(msg):
                nonlocal message
                message = msg
                sleep(1)
                skill.bus.emit(Message(msg.data.get('speak_ident')))
                sleep(1)
                check_for_signal("isSpeaking")
            Thread(target=handler, args=(msg,), daemon=True).start()

        skill = get_test_mycroft_skill(
            {"speak": on_speak})

        # Test wait with speak API
        speak_time = time()
        create_signal("isSpeaking")  # Mock signal create
        skill.speak('test', wait=True)
        # Make sure we actually waited
        self.assertGreaterEqual(time(), speak_time + 1)
        self.assertLessEqual(time(), speak_time + 2)
        self.assertTrue(check_for_signal("neon_speak_api", -1))
        self.assertIsInstance(message, Message)
        self.assertIsInstance(message.data['speak_ident'], str)

        wait_for_signal_clear('isSpeaking')  # Wait for first test to finish
        # Test wait with signals
        self.assertTrue(check_for_signal("neon_speak_api"))
        speak_time = time()
        create_signal("isSpeaking")  # Mock signal create
        skill.speak('test', wait=True)
        # Make sure we actually waited
        self.assertGreaterEqual(time(), speak_time + 2)
        self.assertFalse(check_for_signal("neon_speak_api", -1))

    # TODO: Test settings load


class NeonSkillTests(unittest.TestCase):
    skill = None
    config_dir = os.path.join(os.path.dirname(__file__), "skills", "config")

    @classmethod
    def setUpClass(cls) -> None:
        from skills.test_skill import TestSkill

        bus = FakeBus()
        os.environ["XDG_CONFIG_HOME"] = cls.config_dir

        cls.skill = TestSkill()
        # Mock the skill_loader process
        if hasattr(cls.skill, "_startup"):
            cls.skill._startup(bus)
        else:
            cls.skill.bind(bus)
            cls.skill.load_data_files()
            cls.skill.initialize()

        cls.skill.config_core.setdefault('language', dict())
        cls.skill.config_core['language']['detection_module'] = "libretranslate_detection_plug"
        cls.skill.config_core['language']['translation_module'] = "libretranslate_plug"

    @classmethod
    def tearDownClass(cls) -> None:
        if os.path.isdir(cls.config_dir):
            shutil.rmtree(cls.config_dir)
        os.environ.pop("XDG_CONFIG_HOME")

    def test_00_skill_init(self):
        self.assertIsInstance(self.skill.cache_loc, str)
        self.assertTrue(os.path.isdir(self.skill.cache_loc))
        self.assertIsNotNone(self.skill.lru_cache)
        self.assertIsInstance(self.skill.sys_tz, datetime.tzinfo)
        self.assertIsInstance(self.skill.neon_core, bool)
        self.assertIsInstance(self.skill.skill_mode, str)
        self.assertIsInstance(self.skill.extension_time, int)
        self.assertEqual(self.skill.config_core['language']['detection_module'],
                         "libretranslate_detection_plug")
        self.assertEqual(self.skill.config_core['language']['translation_module'],
                         "libretranslate_plug")
        self.assertIsNotNone(self.skill.lang_detector)
        self.assertIsNotNone(self.skill.translator)

    def test_properties(self):
        self.assertIsInstance(self.skill.gui_enabled, bool)
        self.assertIsInstance(self.skill.ngi_settings, dict)
        self.assertEqual(self.skill.ngi_settings, self.skill.settings)

    def test_preference_skill(self):
        self.assertIsInstance(self.skill.preference_skill(), dict)
        self.assertEqual(self.skill.preference_skill()["boolean_type"], False)
        self.assertEqual(self.skill.preference_skill()["number_type"], 2.0)
        self.assertEqual(self.skill.preference_skill()["text_type"], "8")
        self.skill.settings['text_type'] = 'test'
        self.assertEqual(self.skill.preference_skill()["text_type"], "test")
        self.skill.settings['new_setting'] = 'test'
        self.assertEqual(self.skill.preference_skill()["new_setting"], "test")

    def test_update_profile(self):
        from neon_utils.configuration_utils import get_neon_user_config
        test_config_path = os.path.dirname(__file__)
        profile_settings = get_neon_user_config(test_config_path)
        test_username = "tester"
        profile_settings['user']['username'] = test_username
        test_message_old = Message("", {},
                                   {'username': test_username,
                                    'nick_profiles': {
                                        test_username: profile_settings.content}})
        test_message_new = Message("", {},
                                   {'username': test_username,
                                    'user_profiles': [profile_settings.content]})
        new_email = "new@email.test"
        self.skill.update_profile({'user': {'email': new_email}},
                                  test_message_old)
        self.assertEqual(test_message_old.context['nick_profiles']
                         [test_username]['user']['email'], new_email)

        self.skill.update_profile({'user': {'email': new_email}},
                                  test_message_new)
        self.assertEqual(test_message_new.context['user_profiles']
                         [0]['user']['email'], new_email)
        os.remove(profile_settings.file_path)
        # TODO: Define and test persistent data

    def test_update_skill_settings(self):
        self.skill.server = False
        settings = self.skill.settings
        self.skill.update_skill_settings(
            {"boolean_type": settings.get("boolean_type")})
        self.assertEqual(settings, self.skill.settings)

        test_val = "updated value test"
        self.skill.update_skill_settings({"text_type": test_val})
        self.assertEqual(self.skill.settings['text_type'], test_val)

        self.skill.update_skill_settings({"text_type": test_val,
                                          "new_pref": True})
        self.assertTrue(self.skill.settings['new_pref'])
        # TODO: Define and test for user-specific settings

    def test_neon_must_respond(self):
        self.assertFalse(self.skill.neon_must_respond())
        private_message_solo = Message("", {},
                                       {"klat_data": {
                                           "title": "!PRIVATE:user"}})
        private_message_neon = Message("", {},
                                       {"klat_data": {
                                           "title": "!PRIVATE:user,Neon"}})
        private_message_neon_plus = Message("", {},
                                            {"klat_data": {
                                                "title": "!PRIVATE:user,Neon,user1"}})
        public_message = Message("", {},
                                 {"klat_data": {
                                     "title": "Test Conversation"}})
        first_message = Message("",
                                {"utterance": "Welcome to your private conversation with Neon"},
                                {"klat_data": {
                                    "title": "!PRIVATE:user"}})
        self.assertFalse(self.skill.neon_must_respond())
        self.assertTrue(self.skill.neon_must_respond(private_message_solo))
        self.assertTrue(self.skill.neon_must_respond(private_message_neon))
        self.assertFalse(self.skill.neon_must_respond(private_message_neon_plus))
        self.assertFalse(self.skill.neon_must_respond(public_message))
        self.assertFalse(self.skill.neon_must_respond(first_message))

    def test_neon_in_request(self):
        # TODO: Mock `is_neon_core` and test `skill.neon_in_request` directly
        from neon_utils.message_utils import request_for_neon

        # Test message context/vocab
        neon_should_respond = Message("test_neon_should_respond", {},
                                      {'neon_should_respond': True})
        self.assertTrue(request_for_neon(neon_should_respond, "neon",
                                         self.skill.voc_match, False))

        neon_in_data = Message("test_neon_should_respond", {'neon': "Neon"},
                               {'neon_should_respond': False})
        self.assertTrue(request_for_neon(neon_in_data, "neon",
                                         self.skill.voc_match, False))

        # Test Config WW state
        self.assertFalse(request_for_neon(Message("test"), "neon",
                                          self.skill.voc_match, False))
        self.assertTrue(request_for_neon(Message("test"), "neon",
                                         self.skill.voc_match, True))

        # Test vocab match
        neon_in_utterance = Message("test_neon_in_utterance",
                                    {'utterance': "hello Neon"},
                                    {"neon_should_respond": False})
        self.assertTrue(request_for_neon(neon_in_utterance, "neon",
                                         self.skill.voc_match, False))

    def test_report_metric(self):
        metric_handler = Mock()
        self.skill.bus.on("neon.metric", metric_handler)
        self.skill.report_metric("test metric", {"name": "invalid name",
                                                 "param": "value",
                                                 "test": True})
        metric_handler.assert_called_once()
        message = metric_handler.call_args[0][0]
        self.assertEqual(message.msg_type, "neon.metric")
        self.assertEqual(message.data, {"name": "test metric",
                                        "param": "value",
                                        "test": True})

    @pytest.mark.skip
    def test_send_email(self):
        self.assertTrue(self.skill.send_email(
            "Test Message",
            "This is a test\ncalled from neon_skill_tests.py in neon-utils",
            email_addr="test@neongecko.com"))

    def test_make_active(self):
        active_request = Mock()
        self.skill.bus.on("active_skill_request", active_request)
        self.skill.make_active(15)
        active_request.assert_called_once()
        message = active_request.call_args[0][0]
        self.assertEqual(message.msg_type, "active_skill_request")
        self.assertEqual(message.data, {"skill_id": self.skill.skill_id,
                                        "timeout": 15})

    def test_request_check_timeout(self):
        set_timeout = Mock()
        self.skill.bus.on("set_timeout", set_timeout)

        self.skill.request_check_timeout(10, "timeout.intent")
        set_timeout.assert_called_once()
        message = set_timeout.call_args[0][0]
        self.assertEqual(message.msg_type, "set_timeout")
        self.assertEqual(message.data,
                         {"time_out": 10,
                          "intent_to_check":
                              f"{self.skill.skill_id}:timeout.intent"})

        set_timeout.reset_mock()
        self.skill.request_check_timeout(30, ["test_intent_1", f"test_intent"])
        self.assertEqual(set_timeout.call_count, 2)

    def test_decorate_api_call_use_lru(self):
        # TODO
        pass


class SkillGuiTests(unittest.TestCase):
    from neon_utils.skills.skill_gui import SkillGUI
    skill = get_test_neon_skill({})
    skill_gui = SkillGUI(skill)

    @classmethod
    def setUpClass(cls) -> None:

        cls.skill_gui.config = {
            "remote-server": None
        }

    def test_pages2uri(self):
        def _find_resource(*args, **kwargs):
            return join("skill_id", args[1], args[0])

        real_method = self.skill.find_resource
        self.skill.find_resource = _find_resource

        self.skill_gui.serving_http = False
        urls = self.skill_gui._pages2uri(["test_page1", "test_page2"])
        self.assertEqual(urls, ["file://skill_id/ui/test_page1",
                                "file://skill_id/ui/test_page2"])

        self.skill_gui.config = {
            "remote-server": "remote_url"
        }
        urls = self.skill_gui._pages2uri(["test_page1", "test_page2"])
        self.assertEqual(urls, ["remote_url/skill_id/ui/test_page1",
                                "remote_url/skill_id/ui/test_page2"])

        self.skill_gui.serving_http = True
        urls = self.skill_gui._pages2uri(["SYSTEM_Test1", "SYSTEM_Test2"])
        self.assertEqual(urls, ["file://remote_url/system/ui/SYSTEM_Test1",
                                "file://remote_url/system/ui/SYSTEM_Test2"])

        urls = self.skill_gui._pages2uri(["test_page1", "test_page2"])
        self.assertEqual(urls, ["file://remote_url/skills/skill_id/ui/test_page1",
                                "file://remote_url/skills/skill_id/ui/test_page2"])

        self.skill.find_resource = real_method


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import glob
import os
import base64
import wave

from tempfile import mkstemp

from typing import Optional, List
from ovos_utils.signal import ensure_directory_exists

from neon_utils.logger import LOG


def encode_file_to_base64_string(path: str) -> str:
    """
    Encodes a file to a base64 string (useful for passing file data over a messagebus)
    :param path: Path to file to be encoded
    :return: encoded string
    """
    if not isinstance(path, str):
        raise TypeError
    path = os.path.expanduser(path)
    if not os.path.isfile(path):
        LOG.error(f"File Not Found: {path}")
        raise FileNotFoundError
    with open(path, "rb") as file_in:
        encoded = base64.b64encode(file_in.read()).decode("utf-8")
    return encoded


def decode_base64_string_to_file(encoded_string: str, output_path: str) -> str:
    """
    Writes out a base64 string to a file object at the specified path
    :param encoded_string: Base64 encoded string
    :param output_path: Path to file to write (throws exception if file exists)
    :return: Path to output file
    """
    if not isinstance(output_path, str):
        raise TypeError
    output_path = os.path.expanduser(output_path)
    if os.path.isfile(output_path):
        LOG.error(f"File already exists: {output_path}")
        raise FileExistsError
    ensure_directory_exists(os.path.dirname(output_path))
    with open(output_path, "wb+") as file_out:
        byte_data = base64.b64decode(encoded_string.encode("utf-8"))
        file_out.write(byte_data)
    return output_path


def get_most_recent_file_in_dir(path: str, ext: Optional[str] = None) -> Optional[str]:
    """
    Gets the most recently created file in the specified path
    :param path: File path or glob pattern to check
    :param ext: File extension
    :return: Path to newest file in specified path with specified extension (None if no files in path)
    """
    if os.path.isdir(path):
        path = f"{path}/*"
    list_of_files = glob.glob(path)  # * means all if need specific format then *.csv
    if ext:
        if not ext.startswith("."):
            ext = f".{ext}"
        list_of_files = [file for file in list_of_files if os.path.splitext(file)[1] == ext]
    if list_of_files:
        latest_file = max(list_of_files, key=os.path.getctime)
        return latest_file
    else:
        return None


def get_file_as_wav(audio_file: str, desired_sample_rate: int, desired_sample_width: int = 2) -> wave.Wave_read:
    """
    Gets a wav file for the passed audio file.
    Args:
        audio_file: Path to audio file in arbitrary format
        desired_sample_rate: sample rate at which returned wav data should be encoded
        desired_sample_width: sample width at which returned wav data should be encoded
    Returns:
        Wave_read object encoded at the desired_sample_rate
    """
    from pydub import AudioSegment
    audio_file = os.path.expanduser(audio_file)
    if not os.path.isfile(audio_file):
        raise FileNotFoundError
    try:
        file = wave.open(audio_file, 'rb')
        sample_rate = file.getframerate()
        sample_width = file.getsampwidth()
        if sample_rate == desired_sample_rate and sample_width == desired_sample_width:
            return file
    except wave.Error:
        pass
    except Exception as e:
        raise e
    audio = AudioSegment.from_file(audio_file)
    audio = audio.set_frame_rate(desired_sample_rate).set_sample_width(desired_sample_width)
    _, tempfile = mkstemp()
    audio.export(tempfile, format='wav').close()
    return wave.open(tempfile, 'rb')


def get_audio_file_stream(wav_file: str, sample_rate: int = 16000):
    """
    Creates a FileStream object for the specified wav_file with the specified output sample_rate.
    Args:
        wav_file: Path to file to read
        sample_rate: Desired output sample rate (None for wav_file sample rate)

    Returns:
        FileStream object for the passed audio file
    """
    class FileStream:
        MIN_S_TO_DEBUG = 5.0

        # How long between printing debug info to screen
        UPDATE_INTERVAL_S = 1.0

        def __init__(self, file_name):
            self.file = get_file_as_wav(file_name, sample_rate)
            self.sample_rate = self.file.getframerate()
            # if sample_rate and self.sample_rate != sample_rate:
            #     sound = AudioSegment.from_file(file_name, format='wav', frame_rate=self.sample_rate)
            #     sound = sound.set_frame_rate(sample_rate)
            #     _, tempfile = mkstemp()
            #     sound.export(tempfile, format='wav')
            #     self.file = wave.open(tempfile, 'rb')
            #     self.sample_rate = self.file.getframerate()
            self.size = self.file.getnframes()
            self.sample_width = self.file.getsampwidth()
            self.last_update_time = 0.0

            self.total_s = self.size / self.sample_rate / self.sample_width

        def calc_progress(self):
            return float(self.file.tell()) / self.size

        def read(self, chunk_size):

            progress = self.calc_progress()
            if progress == 1.0:
                raise EOFError

            return self.file.readframes(chunk_size)

        def close(self):
            self.file.close()

    if not os.path.isfile(wav_file):
        raise FileNotFoundError
    try:
        return FileStream(wav_file)
    except Exception as e:
        raise e


def audio_bytes_from_file(file_path: str) -> dict:
    """
        :param file_path: Path to file to read

        :returns {sample_rate, data, audio_format} if audio format is supported, empty dict otherwise
    """
    try:
        import librosa
    except ImportError:
        raise ImportError("librosa not available,"
                          " pip install neon-utils[audio]")
    except OSError as e:
        if repr(e) == "sndfile library not found":
            raise OSError("libsndfile missing, install via: "
                          "sudo apt install libsndfile1")
        else:
            raise e

    supported_audio_formats = ['mp3', 'wav']
    audio_format = file_path.split('.')[-1]
    if audio_format in supported_audio_formats:
        data, sample_rate = librosa.load(file_path)
        raw_audio_data = dict(sample_rate=sample_rate,
                              data=data,
                              audio_format=audio_format)
    else:
        LOG.warning(f'Failed to resolve audio format: {audio_format}')
        raw_audio_data = dict()

    return raw_audio_data


def audio_bytes_to_file(file_path: str, audio_data: List[float], sample_rate: int) -> str:
    """
        :param file_path: Path to file to write
        :param audio_data: array of audio data as float time series
        :param sample_rate: audio data sample rate

        :returns Path to saved file if saved successfully None otherwise
    """
    try:
        import soundfile as sf
    except ImportError:
        raise ImportError("soundfile not available,"
                          " pip install neon-utils[audio]")
    try:
        sf.write(file=file_path, data=audio_data, samplerate=sample_rate)
    except Exception as ex:
        LOG.error(f'Exception occurred while writing '
                  f'{audio_data} to {file_path}: {ex}')
        file_path = None
    return file_path


def resolve_neon_resource_file(res_name: str) -> Optional[str]:
    """
    Locates a resource file bundled with neon_utils or neon_core
    :param res_name: resource name (i.e. snd/start_listening.wav) to locate
    :return: path to resource or None if resource is not found
    """
    base_dir = os.path.join(os.path.dirname(__file__), "res")
    res_file = os.path.join(base_dir, res_name)
    if os.path.isfile(res_file):
        return res_file

    from neon_utils.packaging_utils import get_neon_core_root
    try:
        base_dir = os.path.join(get_neon_core_root(), "res")
    except FileNotFoundError:
        LOG.warning("No neon_core directory found")
        return None

    res_file = os.path.join(base_dir, res_name)
    if os.path.isfile(res_file):
        return res_file
    LOG.warning(f"Requested res_file not found: {res_file}")
    return None


def parse_skill_readme_file(readme_path: str) -> dict:
    """
    Parse a Neon skill README.md into a dictionary
    :param readme_path: absolute path to README.md file
    :return: dict of parsed data
    """
    from neon_utils.parse_utils import clean_quotes

    # Check passed path
    if not readme_path:
        raise ValueError("Null path")
    readme_path = os.path.expanduser(readme_path)
    if not os.path.isfile(readme_path):
        raise FileNotFoundError(f"{readme_path} is not a valid file")

    # Initialize parser params
    list_sections = ("examples", "incompatible skills", "platforms",
                     "categories", "tags", "credits")
    section = "header"
    category = None
    parsed_data = {}
    with open(readme_path, "r") as readme:
        lines = readme.readlines()

    def _check_section_start(ln: str):
        # Handle section start
        if ln.startswith("# ![](https://0000.us/klatchat/app/files/"
                         "neon_images/icons/neon_paw.png)"):
            # Old style title line
            parsed_data["title"] = ln.split(')', 1)[1].strip()
            parsed_data["icon"] = ln.split('(', 1)[1].split(')', 1)[0].strip()
            return
        elif section == "header" and ln.startswith("# <img src="):
            # Title line
            parsed_data["title"] = ln.split(">", 1)[1].strip()
            parsed_data["icon"] = ln.split("src=", 1)[1].split()[0]\
                .strip('"').strip("'").lstrip("./")
            return "summary"
        elif ln.startswith("# ") or ln.startswith("## "):
            # Top-level section
            if ln.startswith("## About"):
                # Handle Mycroft 'About' section
                return "description"
            elif ln.startswith("## Category"):
                # Handle 'Category' as 'Categories'
                return "categories"
            else:
                return line.lstrip("#").strip().lower()
        return

    def _format_readme_line(ln: str):
        nonlocal category
        if section == "incompatible skills":
            if not any((ln.startswith('-'), ln.startswith('*'))):
                return None
            parsed = clean_quotes(ln.lstrip('-').lstrip('*').lower().strip())
            if parsed.startswith('['):
                return parsed.split('(', 1)[1].split(')', 1)[0]
            return parsed
        if section == "examples":
            if not any((ln.startswith('-'), ln.startswith('*'))):
                return None
            parsed = clean_quotes(ln.lstrip('-').lstrip('*').strip())
            if parsed.split(maxsplit=1)[0].lower() == "neon":
                return parsed.split(maxsplit=1)[1]
            else:
                return parsed
        if section == "categories":
            parsed = ln.rstrip('\n').strip('*')
            if ln.startswith('**'):
                category = parsed
            return parsed
        if section == "credits":
            if ln.strip().startswith('['):
                return ln.split('[', 1)[1].split(']', 1)[0]
            return ln.rstrip('\n').lstrip('@')
        if section == "tags":
            return ln.lstrip('#').rstrip('\n')
        if section in list_sections:
            return clean_quotes(ln.lstrip('-').lstrip('*').lower().strip())
        return ln.rstrip('\n').rstrip()

    for line in lines:
        new_section = _check_section_start(line)
        if new_section:
            section = new_section
        elif line.strip():
            parsed_line = _format_readme_line(line)
            if not parsed_line:
                # Nothing to parse in this line
                continue
            if section in list_sections:
                if section not in parsed_data:
                    parsed_data[section] = list()
                parsed_data[section].append(parsed_line)
            else:
                if section not in parsed_data:
                    parsed_data[section] = parsed_line
                else:
                    parsed_data[section] = " ".join((parsed_data[section],
                                                     parsed_line))
    parsed_data["category"] = category or parsed_data.get("categories",
                                                          [""])[0]
    if parsed_data.get("incompatible skills"):
        parsed_data["incompatible_skills"] = \
            parsed_data.pop("incompatible skills")
    if parsed_data.get("credits") and len(parsed_data["credits"]) == 1:
        parsed_data["credits"] = parsed_data["credits"][0].split(' ')

    return parsed_data


def check_path_permissions(path: str) -> tuple:
    """
    Check current access permissions for the specified path
    :param path: file or directory path to check
    :return: tuple (read, write, execute)
    """
    path = os.path.expanduser(path)
    if not os.access(path, os.F_OK):
        raise FileNotFoundError(f"{path} does not exist")
    stat = (os.access(path, os.R_OK), os.access(path, os.W_OK),
            os.access(path, os.X_OK))
    return stat


def path_is_read_writable(path: str) -> bool:
    """
    Check if the specified path is readable and writable by the current user.
    :param path: str file path to check
    :return: True if path exists and is read/writable
    """
    if not path:
        return False
    try:
        path = os.path.expanduser(path)
        while path and not os.path.exists(path):
            path = os.path.dirname(path)
        stat = check_path_permissions(path)
        return all(stat[:2])
    except FileNotFoundError:
        return False


def create_file(filename: str):
    """
    Create the file filename and create any directories needed
    :param filename: Path to the file to be created
    """
    filename = os.path.expanduser(filename)
    if not path_is_read_writable(os.path.dirname(filename)):
        raise PermissionError(f"Insufficient permissions to create {filename}")
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    # with NamedLock(filename): # TODO: Implement combo_lock with file lock support or add lock utils to neon_utils DM
    with open(filename, 'w') as f:
        f.write('')


def load_commented_file(filename: str, comment_prefix: str = '#') -> str:
    """
    Load a text file removing commented lines and trailing newlines
    :param filename: Path to the file to be created
    :param comment_prefix: Prefix characters to count as comments
    """
    # Check passed path
    if not filename:
        raise ValueError("Null path")
    filename = os.path.expanduser(filename)
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"{filename} is not a valid file")

    with open(filename) as f:
        lines = f.readlines()

    valid_lines = [line for line in lines if not
                   line.strip().startswith(comment_prefix)]

    return "".join(valid_lines).rstrip('\n')



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import tracemalloc

from threading import Event, Thread
from typing import Optional
from ovos_utils.log import LOG


_malloc_event = Event()
_malloc_thread = None


def start_systemd_service(service: callable, **kwargs):
    """
    Start a Neon Core module with systemd wrappers to report process lifecycle
    """
    try:
        import sdnotify
    except ImportError:
        LOG.exception(f'sdnotify not installed! '
                      f'Starting service without systemd hooks')
        service(**kwargs)
        return
    notifier = sdnotify.SystemdNotifier()

    def on_ready():
        notifier.notify('READY=1')
        notifier.notify('STATUS=Ready')

    def on_stopping():
        try:
            stop_malloc()
        except Exception as e:
            LOG.error(e)
        notifier.notify('STOPPING=1')
        notifier.notify('STATUS=Stopping')

    def on_error(err: str):
        if err.isnumeric():
            notifier.notify(f'ERRNO={err}')
        else:
            notifier.notify(f'ERRNO=1')

    def on_alive():
        notifier.notify('STATUS=Starting')

    def on_started():
        notifier.notify('STATUS=Started')

    service(ready_hook=on_ready, error_hook=on_error,
            stopping_hook=on_stopping, alive_hook=on_alive,
            started_hook=on_started, **kwargs)


def start_malloc(config: dict = None, stack_depth: int = 1,
                 force: bool = False) -> bool:
    """
    Start malloc trace if configured
    :param config: dict Configuration
    :param stack_depth: depth to track memory usage
    :param force: if True, start tracemalloc regardless of configuration
    :returns: True if malloc started
    """
    if not config:
        from ovos_config.config import Configuration
        config = Configuration()
    if force or config.get('debugging') and \
            config['debugging'].get('tracemalloc'):
        LOG.info(f"starting tracemalloc")
        tracemalloc.start(stack_depth)
        if config['debugging'].get('log_malloc'):
            interval_minutes = config['debugging'].get('log_interval_minutes',
                                                       60)
            global _malloc_thread
            _malloc_thread = Thread(target=_log_malloc,
                                    args=((interval_minutes * 60),),
                                    daemon=True)
            _malloc_thread.start()
        return True
    return False


def stop_malloc():
    """
    Stop tracemalloc logging if active
    """
    if _malloc_thread:
        LOG.debug(f"Stopping malloc logging")
        _malloc_event.set()
        _malloc_thread.join()


def snapshot_malloc() -> Optional[tracemalloc.Snapshot]:
    """
    Capture and return a tracemalloc Snapshot if tracemalloc is started
    """
    try:
        LOG.debug("Capturing malloc snapshot")
        return tracemalloc.take_snapshot()
    except RuntimeError:
        LOG.debug("No tracemalloc trace")
    except Exception as e:
        LOG.exception(e)
    return None


def print_malloc(snapshot: tracemalloc.Snapshot, limit: int = 8,
                 filter_traces: bool = False):
    """
    Log a malloc snapshot
    :param snapshot: Snapshot to evaluate
    :param limit: number of traces to log
    :param filter_traces: if True, remove importlib and unknown stack traces
    """
    LOG.debug(f"Processing snapshot")
    if filter_traces:
        snapshot = snapshot.filter_traces((
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<frozen importlib._bootstrap_external>"),
            tracemalloc.Filter(False, "<unknown>"),
        ))
    top_stats = snapshot.statistics('traceback')

    LOG.info(f"Top {limit} memory users")
    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]
        LOG.info(f"#{index}: {frame.filename}:{frame.lineno}: "
                 f"{stat.size / 1048576} MiB")
        for frame in stat.traceback[1:]:
            LOG.debug(f'{frame.filename}:{frame.lineno}')

    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        LOG.info(f"{len(other)} other:{size / 1048576} MiB")
    total = sum(stat.size for stat in top_stats)
    LOG.info(f"Total allocated size: {total / 1048576} MiB")


def _log_malloc(interval_seconds: int):
    _malloc_event.clear()
    while not _malloc_event.wait(interval_seconds):
        print_malloc(snapshot_malloc(), filter_traces=True)
    LOG.info(f"Stopping malloc logging")



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from time import sleep

import sys
import os
import unittest

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from neon_utils.metrics_utils import *


class MetricUtilTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        os.environ["NEON_CONFIG_PATH"] = os.path.join(os.path.dirname(__file__), "configuration")

    @classmethod
    def tearDownClass(cls) -> None:
        os.environ.pop("NEON_CONFIG_PATH")

    def test_stopwatch_simple(self):
        sleep_time = 1.00
        stopwatch = Stopwatch()
        with stopwatch:
            sleep(sleep_time)
        self.assertEqual(round(stopwatch.time, 2), sleep_time)

    def test_stopwatch_reuse(self):
        sleep_time = 0.5
        stopwatch = Stopwatch()
        with stopwatch:
            sleep(sleep_time)
        self.assertEqual(round(stopwatch.time, 2), sleep_time)

        with stopwatch:
            sleep(sleep_time)
        self.assertEqual(round(stopwatch.time, 2), sleep_time)

        with stopwatch:
            sleep(sleep_time)
        self.assertEqual(round(stopwatch.time, 2), sleep_time)

    def test_stopwatch_init_params(self):
        stopwatch = Stopwatch("Test", True)
        self.assertEqual(stopwatch._metric, "Test")
        self.assertTrue(stopwatch._report)

        stopwatch = Stopwatch()
        self.assertIsNone(stopwatch._metric)
        self.assertFalse(stopwatch._report)

    def test_report_metric(self):
        self.assertTrue(report_metric("test", data="this is only a test"))

    def test_announce_connection(self):
        # TODO: Better test to check call to `send_mq_request`
        self.assertIsInstance(announce_connection(), bool)


if __name__ == '__main__':
    unittest.main()
